#include "emitter.h"
#include "gmath.h"
#include "texturemanager.h"
#include "singleton.h"
#include "draw.h"
#include "input.h"
#include "collision.h"

Emitter::Emitter(int maxParticles, float life, int emitRadius, const Vector &origin, float angle, int velocity) : ParticleSystem(maxParticles, life, origin)
{
	m_emit = 0.0f;
	m_emitRadius = emitRadius;
	m_velocity = velocity;
	m_emitAngle = angle;
	m_emitDir = AngleToVector(angle, m_emitRadius, m_velocity);
	m_active = false;
};

Emitter::~Emitter()
{
};

void Emitter::UpdatePos(Vector &origin, Vector &emitDir)
{
	m_origin = origin;
	m_emitDir = emitDir;
};

void Emitter::ProcessInput()
{
	int mouseX = Input::GetSingleton().GetX();
	int mouseY = 600 - Input::GetSingleton().GetY();
	int oldX = Input::GetSingleton().GetOldX();
	int oldY = Input::GetSingleton().GetOldY();
	if(InCircle(Vector(mouseX, mouseY, 0), m_origin, m_emitRadius) == true)
	{
		if(Input::GetSingleton().mouseStillDown(SDL_BUTTON_LEFT) == true)
		{
			if(oldX != mouseX || oldY != mouseY)
			{
				m_origin.x = mouseX;
				m_origin.y = mouseY;
			}
		}
		else if(Input::GetSingleton().mouseStillDown(SDL_BUTTON_RIGHT) == true)
		{
			if(oldX != mouseX)
			{
				if(mouseX > oldX)
				{
					m_emitRadius += (mouseX - oldX);
					if(m_emitRadius > 100)
						m_emitRadius = 100;
				}
				else if(mouseX < oldX)
				{
					m_emitRadius -= (oldX - mouseX);
					if(m_emitRadius < 20)
						m_emitRadius = 20;
				}
			}
		}
		else if(Input::GetSingleton().mouseDown(SDL_BUTTON_WHEELDOWN) == true)
		{
			m_emitAngle += 1;
			if(m_emitAngle >= 360)
				m_emitAngle = 0;
			m_emitDir = AngleToVector(m_emitAngle, m_emitRadius, m_velocity);
		}
		else if(Input::GetSingleton().mouseDown(SDL_BUTTON_WHEELUP) == true)
		{
			m_emitAngle -= 1;
			if(m_emitAngle < 0)
				m_emitAngle = 360;
			m_emitDir = AngleToVector(m_emitAngle, m_emitRadius, m_velocity);
		}
		m_active = true;
	}
	else
		m_active = false;

};

void Emitter::Update(float timems)
{
	m_accumulatedTime += timems;
	for(int i = 0; i < m_maxParticles; i++)
	{
		m_particleList[i].Update(timems);

		if(!m_particleList[i].IsAlive())
			m_numParticles--;
	}
};

void Emitter::Render(bool reflect)
{
	if(m_active == true)
		DrawLineCirle(m_origin, m_emitRadius, COLOUR_WHITE);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_SMOOTH);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glDepthMask(GL_FALSE);
	m_texture->Bind();
	float x, y;
	x = m_texture->GetWidth() / 2;
	y = m_texture->GetHeight() / 2;
	glBegin(GL_QUADS);
	{
		for(int i = 0; i < m_maxParticles; i++)
		{
			if(m_particleList[i].IsAlive())
				m_particleList[i].DrawQuad(x, y);
		}
	}
	glEnd();
};

int Emitter::Emit(float numParticles)
{
	if(m_emitting == false)
		return 0;
	m_emit += numParticles;

	int num_particles = 0;
	int emitted = 0;
	int lastFreeSlot = 0;

	if(m_emit > 1.0)
	{
		num_particles = static_cast<int>(m_emit);
		m_emit -= num_particles;
	}

	while(num_particles && (m_numParticles < m_maxParticles))
	{
		for(int i = lastFreeSlot; i < m_maxParticles; i++)
		{
			if(!m_particleList[i].IsAlive())
			{
				InitParticle(m_particleList[i]);
				m_numParticles++;
				num_particles--;
				emitted--;
				lastFreeSlot = i + 1;
				break;
			}
		}
	}
	return emitted;
};

void Emitter::LoadTexture(std::string filename)
{
	m_texture = TextureManager::GetSingleton().Load(filename);
};

void Emitter::InitParticle(Particle &p)
{
	Vector origin;
	origin.x = m_origin.x + (rand() % m_emitRadius);
	origin.y = m_origin.y + (rand() % m_emitRadius);
	origin.z = m_origin.z;
	p.Init(origin, Vector(m_emitDir.x, m_emitDir.y, 0), Vector(0,0,0), m_particleLife, 0.2f, 1.0f, COLOUR_WHITE, WHITE);


};
