#include "image.h"
#include "log.h"

ImageBuffer::ImageBuffer(const std::string &filename) : m_filename(filename), m_data(NULL), m_length(0)
{
	std::ifstream ifs(filename.c_str(), std::ios::in | std::ios::binary);
	if(ifs.fail())
		throw ImageException("Couldn't open file", filename);

	ifs.seekg(0, std::ios::end);
	m_length = ifs.tellg();
	ifs.seekg(0, std::ios::beg);

	try
	{
		m_data = new GLubyte[m_length];

		ifs.read(reinterpret_cast<char *>(m_data), m_length);
		ifs.close();
	}
	catch(...)
	{
		delete[] m_data;
		throw;
	}
};

ImageBuffer::ImageBuffer(GLubyte *pixels) : m_filename("raw_pixels"), m_data(NULL), m_length(0)
{
	try
	{
		m_length = strlen((char *)pixels);
		m_filename.append((char *)pixels);
		m_data = new GLubyte[m_length];
		m_data = pixels;
	}
	catch(...)
	{
		delete[] m_data;
		throw;
	}
};

ImageBuffer::~ ImageBuffer()
{
	delete[] m_data;
	m_data = NULL;
};

Image::Image(const std::string &name, GLsizei w, GLsizei h, GLint numMipMaps,
		GLenum format, GLint components, const GLubyte *pixels, bool stdCoordSys) : m_width(w), m_height(h),
			m_numMipmaps(numMipMaps), m_format(format), m_components(components), m_name(name),
			m_standerdCoordSystem(stdCoordSys)
{
	long size = m_width * m_height * m_components;

	if(size <= 0)
		throw std::invalid_argument("Image::Image Invalid width, height or component value");

	if(!pixels)
		throw std::invalid_argument("Image::Image Invalid Pixel source");

	m_pixels = new GLubyte[size];

	memcpy(m_pixels, pixels, size);
};

Image::~Image()
{
	delete[] m_pixels;
	m_pixels = NULL;

};

bool Image::IsCompressed() const
{
	switch(m_format)
	{
		case GL_COMPRESSED_RGBA_S3TC_DXT1_EXT:
		case GL_COMPRESSED_RGBA_S3TC_DXT3_EXT:
		case GL_COMPRESSED_RGBA_S3TC_DXT5_EXT:
			return true;
			break;
		default:
			return false;
	}
};

bool Image::isPowerOfTwo() const
{
	GLsizei m;
	for(m = 1; m < m_width; m *= 2)
		;

	if(m != m_width)
		return false;

	if(m != m_height)
		return false;

	return true;
};

template<char ch0, char ch1, char ch2, char ch3>
struct MakeFourCC
{
	enum
	{
		Val = (((ch3 << 24) & 0xFF000000) |
		       ((ch2 << 16) & 0x00FF0000) |
		       ((ch1 <<  8) & 0x0000FF00) |
		       (ch0         & 0x000000FF))
	};
};

const unsigned int FOURCC_DXT1 = MakeFourCC<'D', 'X', 'T', '1'>::Val;
const unsigned int FOURCC_DXT3 = MakeFourCC<'D', 'X', 'T', '3'>::Val;
const unsigned int FOURCC_DXT5= MakeFourCC<'D', 'X', 'T', '5'>::Val;

ImageDDS::ImageDDS(const ImageBuffer &ibuff) : m_ddsd(NULL)
{
	m_name = ibuff.filename();

	m_standerdCoordSystem = false;

	if(!GLEW_ARB_texture_compression)
		throw ImageException("missing GL_ARB_texture_compression extention", m_name);

	try
	{
		const GLubyte *data_ptr = ibuff.data();
		int buffersize = ibuff.length();

		std::string magic;
		magic.assign(reinterpret_cast<const char *>(data_ptr), 4);
		if(magic.compare(0, 4, "DDS ") != 0)
			throw ImageException("Not a DDS file", m_name);

		data_ptr += 4;
		buffersize -= 4;

		m_ddsd = reinterpret_cast<const DDSSurfaceDesc *>(data_ptr);
		data_ptr += sizeof(DDSSurfaceDesc);
		buffersize += sizeof(DDSSurfaceDesc);

		m_width = m_ddsd->width;
		m_height = m_ddsd->height;
		m_numMipmaps = m_ddsd->mipMapLevels;

		switch(m_ddsd->format.fourCC)
		{
			case FOURCC_DXT1:
				m_format = GL_COMPRESSED_RGBA_S3TC_DXT1_EXT;
				m_components = 3;
			break;
			case FOURCC_DXT3:
				m_format = GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
				m_components = 4;
			break;
			case FOURCC_DXT5:
				m_format = GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
				m_components = 4;
			break;

			default:
				throw ImageException("Unsupported DXT format", m_name);
		}

		m_pixels = new GLubyte[buffersize];
		memcpy(m_pixels, data_ptr, buffersize);
	}
	catch(...)
	{
		delete[] m_pixels;
		throw;
	}
};

int ImageTGA::rgbaTable[4] = { 2, 1 ,0 ,3 };
int ImageTGA::bgraTable[4] = { 0, 1, 2, 3 };

ImageTGA::ImageTGA(const ImageBuffer &ibuff)
{
	const GLubyte *colourmap = NULL;
	const GLubyte *data_ptr;

	try
	{
		m_name = ibuff.filename();
		data_ptr = ibuff.data();
		m_standerdCoordSystem = true;

		m_header = reinterpret_cast<const TGAHeader *>(data_ptr);
		data_ptr += sizeof(TGAHeader) + m_header->id_len;
		printf("=======================================\n");
		printf("Image:%s\n", m_name.c_str());
		printf("id_len: %d\n", m_header->id_len);
		printf("width: %d\n", m_header->width);
		printf("height: %d\n", m_header->height);
		printf("colourmapType: %d\n", m_header->colourmap_type);
		printf("Image type:%d\n",m_header->image_type);
		printf("pixel depth: %d\n", m_header->pixel_depth);
		getTextureInfo();

		m_pixels = new GLubyte[m_width * m_height * m_components];

		if(m_header->colourmap_type)
		{
			colourmap = data_ptr;
			data_ptr += m_header->cm_length * (m_header->cm_size >> 3);
		}


		switch(m_header->image_type)
		{
			case 0:
			break;
			case 1:
				ReadTGA8bits(data_ptr, colourmap);
			break;
			case 2:
				switch(m_header->pixel_depth)
				{
					case 16:
						ReadTGA16bits(data_ptr);
					break;
					case 24:
						ReadTGA24bits(data_ptr);
					break;
					case 32:
						ReadTGA32bits(data_ptr);
					break;
				};
			break;
			case 3:
				if(m_header->pixel_depth == 8)
					ReadTGAgray8bits(data_ptr);
				else
					ReadTGAgray16bits(data_ptr);
			break;
			case 9:
				ReadTGA8bitsRLE(data_ptr, colourmap);
			break;
			case 10:
				switch(m_header->pixel_depth)
				{
					case 16:
						ReadTGA16bitsRLE(data_ptr);
					break;
					case 24:
						ReadTGA24bitsRLE(data_ptr);
					break;
					case 32:
						ReadTGA32bitsRLE(data_ptr);
					break;
				}
			break;
			case 11:
				if(m_header->pixel_depth == 8)
					ReadTGAgray8bitsRLE(data_ptr);
				else
					ReadTGAgray8bitsRLE(data_ptr);
			break;

			default:
				throw ImageException("Unknown TGA image Type", m_name);
		};
	}
	catch(...)
	{
		delete[] m_pixels;
		throw;
	}
};

void ImageTGA::getTextureInfo()
{
	m_width = m_header->width;
	m_height = m_header->height;

	switch(m_header->image_type)
	{
		case 3:
		case 11:
		{
			if(m_header->pixel_depth == 8)
			{
				m_format = GL_LUMINANCE;
				m_components = 1;
			}
			else
			{
				m_format = GL_LUMINANCE_ALPHA;
				m_components = 2;
			}
		break;
		}
		case 1:
		case 2:
		case 9:
		case 10:
		{
			if(m_header->pixel_depth <= 24)
			{
				m_format = GLEW_EXT_bgra ? GL_BGR : GL_RGB;
				m_components = 3;
			}
			else
			{
				m_format = GLEW_EXT_bgra ? GL_BGRA : GL_RGBA;
				m_components = 4;
			}
		break;
		}
	};
};

void ImageTGA::ReadTGA8bits(const GLubyte *data, const GLubyte *colourmap)
{
	const GLubyte *pData = data;
	int *compTable = GLEW_EXT_bgra ? bgraTable : rgbaTable;

	for(int i = 0; i < m_width * m_height; i++)
	{
		GLubyte colour = *(pData++);

		m_pixels[(i * 3) + compTable[0]] = colourmap[(colour * 3) + 0];
		m_pixels[(i * 3) + compTable[1]] = colourmap[(colour * 3) + 1];
		m_pixels[(i * 3) + compTable[2]] = colourmap[(colour * 3) + 2];
	}
};

void ImageTGA::ReadTGA16bits(const GLubyte *data)
{
	const GLubyte *pData = data;
	int *compTable = GLEW_EXT_bgra ? bgraTable : rgbaTable;

	for(int i = 0; i < m_width * m_height; i++, pData += 2)
	{
		unsigned short colour = pData[0] + (pData[1] << 8);

		m_pixels[(i * 3) + compTable[2]] = (((colour & 0x7C00) >> 10) << 3);
		m_pixels[(i * 3) + compTable[1]] = (((colour & 0x03E0) >> 5) << 3);
		m_pixels[(i * 3) + compTable[0]] = (((colour & 0x001F) >> 0) << 3);
	}
};

void ImageTGA::ReadTGA24bits(const GLubyte *data)
{
	if(GLEW_EXT_bgra)
		memcpy(m_pixels, data, m_width * m_height * 3);
	else
	{
		const GLubyte *pData = data;

		for(int i = 0; i < m_width * m_height; i++)
		{
			m_pixels[(i * 3) + rgbaTable[0]] = *(pData++);
			m_pixels[(i * 3) + rgbaTable[1]] = *(pData++);
			m_pixels[(i * 3) + rgbaTable[2]] = *(pData++);

		}
	}
};

void ImageTGA::ReadTGA32bits(const GLubyte *data)
{
	if(GLEW_EXT_bgra)
		memcpy(m_pixels, data, m_width * m_height * 4);
	else
	{
		const GLubyte *pData = data;

		for(int i = 0; i < m_width * m_height; i++)
		{
			m_pixels[(i * 4) + rgbaTable[0]] = *(pData++);
			m_pixels[(i * 4) + rgbaTable[1]] = *(pData++);
			m_pixels[(i * 4) + rgbaTable[2]] = *(pData++);
			m_pixels[(i * 4) + rgbaTable[3]] = *(pData++);

		}
	}

};

void ImageTGA::ReadTGAgray8bits(const GLubyte *data)
{
	memcpy(m_pixels, data, m_width * m_height);
};

void ImageTGA::ReadTGAgray16bits(const GLubyte *data)
{
	memcpy(m_pixels, data, m_width * m_height * 2);
};

void ImageTGA::ReadTGA8bitsRLE(const GLubyte *data, const GLubyte *colourmap)
{
	GLubyte *ptr = m_pixels;
	const GLubyte *pData = data;
	int *compTable = GLEW_EXT_bgra ? bgraTable : rgbaTable;

	while(ptr < m_pixels + (m_width * m_height) * 3)
	{
		GLubyte packet_header = *(pData++);
		int size = 1 + (packet_header & 0x7F);

		if(packet_header & 0x80)
		{
			GLubyte colour = *(pData++);

			for(int i = 0; i < size; i++, ptr += 3)
			{
				ptr[0] = colourmap[(colour * 3) + compTable[0]];
				ptr[1] = colourmap[(colour * 3) + compTable[1]];
				ptr[2] = colourmap[(colour * 3) + compTable[2]];

			}
		}
		else
		{
			for(int i = 0; i < size; i++, ptr += 3)
			{
				GLubyte colour = *(pData++);

				ptr[0] = colourmap[(colour * 3) + compTable[0]];
				ptr[1] = colourmap[(colour * 3) + compTable[1]];
				ptr[2] = colourmap[(colour * 3) + compTable[2]];
			}
		}
	}
};

void ImageTGA::ReadTGA16bitsRLE(const GLubyte *data)
{
	GLubyte *ptr = m_pixels;
	const GLubyte *pData = data;
	int *compTable = GLEW_EXT_bgra ? bgraTable : rgbaTable;

	while(ptr < m_pixels + (m_width * m_height) * 3)
	{
		GLubyte packet_header = *(pData++);
		int size = 1 + (packet_header & 0x7F);

		if(packet_header & 0x80)
		{
			unsigned short colour = pData[0] + (pData[1] << 8);

			for(int i = 0; i < size; i++, ptr += 3)
			{
				ptr[compTable[2]] = (((colour & 0x7C00) >> 10) << 3);
				ptr[compTable[1]] = (((colour & 0x03E0) >> 5) << 3);
				ptr[compTable[0]] = (((colour & 0x001F) >> 0) << 3);
			}
		}
		else
		{
			for(int i = 0; i < size; i++, ptr += 3)
			{
				unsigned short colour = pData[0] + (pData[1] << 8);
				pData += 2;

				ptr[compTable[2]] = (((colour & 0x7C00) >> 10) << 3);
				ptr[compTable[1]] = (((colour & 0x03E0) >> 5) << 3);
				ptr[compTable[0]] = (((colour & 0x001F) >> 0) << 3);
			}
		}
	}
};

void ImageTGA::ReadTGA24bitsRLE(const GLubyte *data)
{
	const GLubyte *pData = data;
	GLubyte *ptr = m_pixels;
	int *compTable = GLEW_EXT_bgra ? bgraTable : rgbaTable;

	while(ptr < m_pixels + (m_width * m_height) * 3)
	{
		GLubyte packet_header = *(pData++);
		int size = 1 + (packet_header & 0x7f);

		if(packet_header & 0x80)
		{
			for(int i = 0; i < size; i++, ptr += 3)
			{
			ptr[0] = pData[compTable[0]];
			ptr[1] = pData[compTable[1]];
			ptr[2] = pData[compTable[2]];
			}

		pData += 3;
		}
		else
		{
			for(int i = 0; i < size; i++, ptr += 3)
			{
				ptr[0] = pData[compTable[0]];
				ptr[1] = pData[compTable[1]];
				ptr[2] = pData[compTable[2]];

				pData += 3;
			}
		}
	}
};

void ImageTGA::ReadTGA32bitsRLE(const GLubyte *data)
{
	const GLubyte *pData = data;
	GLubyte *ptr = m_pixels;
	int *compTable = GLEW_EXT_bgra ? bgraTable : rgbaTable;

	while(ptr < m_pixels + (m_width * m_height) * 4)
	{
		GLubyte packet_header = *(pData++);
		int size = 1 + (packet_header & 0x7f);

		if(packet_header & 0x80)
		{
			for(int i = 0; i < size; i++, ptr += 4)
			{
				ptr[0] = pData[compTable[0]];
				ptr[1] = pData[compTable[1]];
				ptr[2] = pData[compTable[2]];
				ptr[3] = pData[compTable[3]];
			}
		pData += 4;
		}
		else
		{
			for(int i = 0; i < size; i++, ptr += 4)
			{
				ptr[0] = pData[compTable[0]];
				ptr[1] = pData[compTable[1]];
				ptr[2] = pData[compTable[2]];
				ptr[3] = pData[compTable[3]];
				pData += 4;
			}
		}
	}
};

void ImageTGA::ReadTGAgray8bitsRLE(const GLubyte *data)
{
	const GLubyte *pData = data;
	GLubyte *ptr = m_pixels;

	while(ptr < m_pixels + (m_width * m_height))
	{
		GLubyte packet_header = *(pData++);
		int size = 1 + (packet_header & 0x7F);

		if(packet_header & 0x80)
		{
			GLubyte colour = *(pData++);

			memset(ptr, colour, size);
			ptr += size;
		}
		else
		{
		memcpy(ptr, pData, size);
		ptr += size;
		pData += size;
		}
	}
};

void ImageTGA::ReadTGAgray16bitsRLE(const GLubyte *data)
{
	const GLubyte *pData = data;
	GLubyte *ptr = m_pixels;

	while(ptr < m_pixels + (m_width * m_height) * 2)
	{
		GLubyte packet_header = *(pData++);
		int size = 1 + (packet_header & 0x7F);

		if(packet_header & 0x80)
		{
			GLubyte colour = *(pData++);
			GLubyte alpha = *(pData++);

			for(int i = 0; i < size; i++, ptr += 2)
			{
				ptr[0] = colour;
				ptr[1] = alpha;
			}
		}
		else
		{
			memcpy(ptr, pData, size * 2);
			ptr += size * 2;
			pData += size * 2;
		}
	}
};

ImagePNG::ImagePNG(const ImageBuffer &ibuff)
{
	png_structp png_ptr = NULL;
	png_infop info_ptr = NULL;
	png_bytep *row_pointers = NULL;
	int bit_depth, colour_type;
	SourceMgr MySourceMgr(ibuff);

	try
	{
		m_name = ibuff.filename();
		m_standerdCoordSystem = true;

		png_byte sig[8];
		memcpy(sig, reinterpret_cast<const png_byte *>(ibuff.data()), 8);

		if(!png_check_sig(sig, 8))
			throw ImageException("Not a valid PNG file", m_name);

		png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, &errorMsg, error_callback, warning_callback);

		if(!png_ptr)
			throw ImageException("Failed to create png read struct", m_name);

		if(setjmp(png_jmpbuf(png_ptr)))
			throw ImageException(errorMsg, m_name);

		info_ptr = png_create_info_struct(png_ptr);

		if(!info_ptr)
			throw ImageException("Failed to create png info struct", m_name);

		png_set_read_fn(png_ptr, &MySourceMgr, read_callback);

		png_read_info(png_ptr, info_ptr);

		bit_depth = png_get_bit_depth(png_ptr, info_ptr);
		colour_type = png_get_color_type(png_ptr, info_ptr);

		if(colour_type == PNG_COLOR_TYPE_PALETTE)
			png_set_palette_to_rgb(png_ptr);

		if(colour_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
			png_set_gray_1_2_4_to_8(png_ptr);

		if(png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
			png_set_tRNS_to_alpha(png_ptr);

		if(bit_depth == 16)
			png_set_strip_16(png_ptr);
		else if(bit_depth < 8)
			png_set_packing(png_ptr);

		png_read_update_info(png_ptr, info_ptr);
		png_uint_32 _width, _height;
		png_get_IHDR(png_ptr, info_ptr, &_width, &_height,
					&bit_depth, &colour_type, NULL, NULL, NULL);

		m_width = _width;
		m_height = _height;

		getTextureInfo(colour_type);

		m_pixels = new GLubyte[m_width * m_height * m_components];

		row_pointers = new png_bytep[m_height];

		for(unsigned int i = 0; i < m_height; i++)
		{
			row_pointers[i] = (png_bytep)(m_pixels + ((m_height - (i + 1)) * m_width * m_components));
			//row_pointers[i] = (png_byte *)malloc(info_ptr->rowbytes);
		}

		png_read_image(png_ptr, row_pointers);
		png_read_end(png_ptr, NULL);
		png_destroy_read_struct(&png_ptr, &info_ptr, NULL);

		delete[] row_pointers;
	}
	catch(...)
	{
		delete [] m_pixels;
		delete [] row_pointers;

		if(png_ptr)
			png_destroy_read_struct(&png_ptr, &info_ptr, NULL);

		throw;
	}
};

void ImagePNG::getTextureInfo(int colour_type)
{
	switch(colour_type)
	{
		case PNG_COLOR_TYPE_GRAY:
			m_format = GL_LUMINANCE;
			m_components = 1;
		break;
		case PNG_COLOR_TYPE_GRAY_ALPHA:
			m_format = GL_LUMINANCE_ALPHA;
			m_components = 2;
		break;
		case PNG_COLOR_TYPE_RGB:
			m_format = GL_RGB;
			m_components = 3;
		break;
		case PNG_COLOR_TYPE_RGB_ALPHA:
			m_format = GL_RGBA;
			m_components = 4;
		break;

		default:
			throw ImageException("Bad PNG colour Type", m_name);
	}
};

void ImagePNG::read_callback(png_structp png_ptr, png_bytep data, png_size_t length)
{
	SourceMgrPtr src = static_cast<SourceMgrPtr>(png_ptr->io_ptr);

	memcpy(data, src->pibuff->data() + src->offset, length);

	src->offset += length;
};

void ImagePNG::error_callback(png_structp png_ptr,  png_const_charp error_msg)
{
	static_cast<std::string *>(png_ptr->error_ptr)->assign(error_msg);

	longjmp(png_jmpbuf(png_ptr), 1);
};

void ImagePNG::warning_callback(png_structp png_ptr, png_const_charp warning_msg)
{
	printf("libpng: %s\n", warning_msg);
	Log::GetSingleton().Write(LOG_APP, "libpng: %s\n", warning_msg);
};
