#ifndef IMAGE_H_
#define IMAGE_H_

#include "globals.h"
#include <png.h>
class ImageException : public std::runtime_error
{
public:
	ImageException(const std::string &err) : std::runtime_error(err){}
	ImageException(const std::string &err, const std::string filename) : std::runtime_error(err), m_which(filename){}
	virtual ~ImageException() throw () {}

	virtual const char *which() const throw()
	{
		return m_which.c_str();
	}
private:
	std::string m_which;
};

class ImageBuffer
{
public:
	ImageBuffer(const std::string &filename);
	ImageBuffer(GLubyte *pixels);
	ImageBuffer(const ImageBuffer &that) : m_filename(that.m_filename), m_data(NULL), m_length(that.m_length)
	{
		m_data = new GLubyte[m_length];
		memcpy(m_data, that.m_data, m_length);
	}
	~ImageBuffer();

	const std::string &filename() const {return m_filename; }
	const GLubyte *data() const { return m_data; }
	size_t length() const { return m_length; }

	ImageBuffer &operator =(const ImageBuffer &rhs)
	{
		this->~ImageBuffer();
		new (this) ImageBuffer(rhs);
		return *this;
	}
private:
	ImageBuffer();
	std::string m_filename;
	GLubyte *m_data;
	size_t m_length;
};

class Image
{
public:
	Image(const std::string &name, GLsizei w, GLsizei h, GLint numMipMaps,
		GLenum format, GLint components, const GLubyte *pixels, bool stdCoordSys);
	virtual ~Image();
	bool IsCompressed() const;
	bool isPowerOfTwo() const;

	GLsizei width() const { return m_width; }
	GLsizei height() const { return m_height; }
	GLint numMipmaps() const { return m_numMipmaps; }
	GLenum format() const { return m_format; }
	GLint components() const { return m_components; }
	const GLubyte *pixels() const { return m_pixels; }
	const std::string &name() const { return m_name; }
	bool stdCoordSys() const { return m_standerdCoordSystem; }

private:
	Image(const Image &img);
protected:
	Image() : m_width(0), m_height(0), m_numMipmaps(0),
		  m_format(0), m_components(0), m_pixels(NULL), m_standerdCoordSystem(true){}

	GLsizei m_width, m_height;
	GLint m_numMipmaps;
	GLenum m_format;
	GLint m_components;
	GLubyte *m_pixels;
	std::string m_name;

	bool m_standerdCoordSystem;
};

class ImageDDS : public Image
{
public:
	ImageDDS(const ImageBuffer &ibuff);
private:
	struct DDSPixelFormat
	{
		GLuint size;
		GLuint flags;
		GLuint fourCC;
		GLuint bpp;
		GLuint redMask;
		GLuint greenMask;
		GLuint blueMask;
		GLuint alphaMask;
	};

	struct DDSCaps
	{
		GLuint caps;
		GLuint caps2;
		GLuint caps3;
		GLuint caps4;
	};

	struct DDSColourKey
	{
		GLuint lowVal;
		GLuint highVal;
	};

	struct DDSSurfaceDesc
	{
		GLuint size;
		GLuint flags;
		GLuint height;
		GLuint width;
		GLuint pitch;
		GLuint depth;
		GLuint mipMapLevels;
		GLuint alphaBitDepth;
		GLuint reserved;
		GLuint surface;

		DDSColourKey ckDestoverlay;
		DDSColourKey ckDestBlt;
		DDSColourKey ckSrcOverlay;
		DDSColourKey ckSrcblt;

		DDSPixelFormat format;
		DDSCaps caps;

		GLuint textureStage;
	};

	const DDSSurfaceDesc *m_ddsd;
};

class ImageTGA : public Image
{
public:
	ImageTGA(const ImageBuffer &ibuff);
private:
	void getTextureInfo();

	void ReadTGA8bits(const GLubyte *data, const GLubyte *colourmap);
	void ReadTGA16bits(const GLubyte *data);
	void ReadTGA24bits(const GLubyte *data );
	void ReadTGA32bits(const GLubyte *data );
	void ReadTGAgray8bits(const GLubyte *data );
	void ReadTGAgray16bits(const GLubyte *data );

	void ReadTGA8bitsRLE(const GLubyte *data, const GLubyte *colourmap);
	void ReadTGA16bitsRLE(const GLubyte *data);
	void ReadTGA24bitsRLE(const GLubyte *data );
	void ReadTGA32bitsRLE(const GLubyte *data );
	void ReadTGAgray8bitsRLE(const GLubyte *data );
	void ReadTGAgray16bitsRLE(const GLubyte *data );

#pragma pack(push, 1)
	struct TGAHeader
	{
		GLubyte id_len;
		GLubyte colourmap_type;
		GLubyte image_type;

		short cm_first_entry;
		short cm_length;
		GLubyte cm_size;

		short x_origin;
		short y_origin;

		short width;
		short height;

		GLubyte pixel_depth;
		GLubyte image_descriptor;
	};
#pragma pack(pop)
	const TGAHeader *m_header;

	static int rgbaTable[4];
	static int bgraTable[4];
};

class ImagePNG : public Image
{
public:
	ImagePNG(const ImageBuffer &ibuff);

	struct SourceMgr
	{
		SourceMgr() : pibuff(NULL), offset(0) {}
		SourceMgr(const ImageBuffer &ibuff) : pibuff(&ibuff), offset(0){}

		const ImageBuffer *pibuff;
		size_t offset;
	};
	typedef SourceMgr *SourceMgrPtr;
private:
	void getTextureInfo(int colour_type);

	static void read_callback(png_structp png_ptr, png_bytep data, png_size_t length);
	static void error_callback(png_structp png_ptr,  png_const_charp error_msg);
	static void warning_callback(png_structp png_ptr, png_const_charp warning_msg);

	std::string errorMsg;
};

class ImageFactory
{
public:
    static Image *CreateImage(const ImageBuffer &ibuff)
    {
        std::string ext;
        Image *result;

        const std::string &filename = ibuff.filename();
        ext.assign(filename, filename.find_last_of('.') + 1, std::string::npos);

        if(ext.compare("dds") == 0)
		result = new ImageDDS(ibuff);
        else if(ext.compare("tga") == 0)
		result = new ImageTGA(ibuff);
	else if(ext.compare("png") == 0)
		result = new ImagePNG(ibuff);
        else
            throw ImageException("Unhandled image file format", filename);

        return result;
    };
};


#endif
