#include "texture.h"
#include "texturemanager.h"
#include "globals.h"
#include "singleton.h"
#include "log.h"

Texture::Texture() : m_handle(0), m_flags(kDefault), m_standerdCoordSystem(true), m_fail(true)
{
	//might wanna check GL errors here
};

Texture::~ Texture()
{
	if(glIsTexture(m_handle))
		glDeleteTextures(1, &m_handle);
	TextureManager::GetSingleton().UnRegisterTexture(name());
};

void Texture::Bind() const
{
	glBindTexture(target(), m_handle);
};

void Texture::Bind(GLenum texUnit) const
{
	glActiveTexture(texUnit);
	glBindTexture(target(), m_handle);
};

GLint Texture::GetCompressionFormat(GLint internalFormat)
{
	if(!GLEW_EXT_texture_compression_s3tc || !GLEW_ARB_texture_compression)
		return internalFormat;

	switch(internalFormat)
	{
		case 1:
			return GL_COMPRESSED_LUMINANCE;
		case 2:
			return GL_COMPRESSED_LUMINANCE_ALPHA;
		case 3:
			return GL_COMPRESSED_RGB;
		case 4:
			return GL_COMPRESSED_RGBA;
		default:
			throw std::invalid_argument("Texture::getCompressionForamt: Bad internal format");
	};
};

GLint Texture::GetInternalFormat(GLint components)
{
	if(m_flags & kCompress)
		return GetCompressionFormat(components);
	else
		return components;
};

Texture2D::Texture2D(const std::string &filename, TextureFlags flags)
{
	try
	{
		ImageBuffer ibuff(filename);
		std::auto_ptr<Image> img(ImageFactory::CreateImage(ibuff));

		create(img.get(), flags);
	}
	catch(std::exception &err)
	{
		printf("Error: Couldn't create texture2D from %s\nReason:%s\n", filename.c_str(), err.what());
		Log::GetSingleton().Write(LOG_APP, "Error: Couldn't create texture2D from %s\nReason:%s\n",
					filename.c_str(), err.what());
	}
};

Texture2D::Texture2D(const Image *img, TextureFlags flags)
{
	try
	{
		create(img, flags);
	}
	catch(std::exception &err)
	{
		printf("Error: Couldn't create texture2D from %s\nReason:%s\n",
					img->name().c_str(), err.what());
		Log::GetSingleton().Write(LOG_APP, "Error: Couldn't create texture2D from %s\nReason:%s\n",
					img->name().c_str(), err.what());
	}
};

void Texture2D::create(const Image *img, TextureFlags flags)
{
	if(!img->pixels())
		throw std::runtime_error("invalid image data");

	m_name = img->name();
	m_flags = flags;
	m_standerdCoordSystem = img->stdCoordSys();
	m_width = img->width();
	m_height = img->height();
	glGenTextures(1, &m_handle);
	glBindTexture(target(), m_handle);
	printf("=====================================\n");
	printf("name: %s\n", img->name().c_str());
	printf("handle: %d\n", m_handle);
	printf("width: %d\n", img->width());
	printf("height %d\n", img->height());
	printf("components: %d\n", img->components());

	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	if(img->IsCompressed())
	{
		GLsizei mipWidth = img->width();
		GLsizei mipHeight = img->height();

		int offset = 0;
		int blockSize = (img->format() == GL_COMPRESSED_RGBA_S3TC_DXT1_EXT) ? 8 : 16;

		for(GLint mipLevel = 0; mipLevel < img->numMipmaps(); mipLevel++)
		{
			GLsizei mipSize = ((mipWidth + 3) / 4) * ((mipHeight + 3) / 4) * blockSize;

			glCompressedTexImage2D(GL_TEXTURE_2D, mipLevel, img->format(), mipWidth, mipHeight, 0,
							mipSize, img->pixels() + offset);

			mipWidth = max(mipWidth >> 1, 1);
			mipHeight = max(mipHeight >> 1, 1);

			offset += mipSize;
		}
	}
	else
	{
		if(GLEW_SGIS_generate_mipmap && img->isPowerOfTwo())
		{
			glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);

			glTexImage2D(GL_TEXTURE_2D, 0, GetInternalFormat(img->components()),
					img->width(), img->height(), 0, img->format(), GL_UNSIGNED_BYTE, img->pixels());
		}
		else
		{
			gluBuild2DMipmaps(GL_TEXTURE_2D, GetInternalFormat(img->components()),
					img->width(), img->height(), img->format(), GL_UNSIGNED_BYTE, img->pixels());
		}
	}

// 	if(glGetError() == GL_NO_ERROR)
		m_fail = false;
};
